@@ -1,25 +1,29 @@
 import hashlib
+import inspect
 import logging
 import re
+from collections.abc import Iterable
+from collections.abc import Iterator
 from typing import Any
-from typing import Iterable
-from typing import Iterator
 from typing import TextIO
-from typing import Union
 from typing import cast
 
 import orjson
 from aiohttp import ClientResponseError
 from humps import main as humps
 from pydantic.dataclasses import dataclass
-from tortoise import fields
 
 from dipdup import env
+from dipdup import fields
 from dipdup.config import DEFAULT_POSTGRES_SCHEMA
 from dipdup.config import HasuraConfig
-from dipdup.config import HTTPConfig
+from dipdup.config import HttpConfig
 from dipdup.config import PostgresDatabaseConfig
-from dipdup.config import ResolvedHTTPConfig
+from dipdup.config import ResolvedHttpConfig
+from dipdup.database import AsyncpgClient
+from dipdup.database import get_connection
+from dipdup.database import iter_models
+from dipdup.database import pg_get_views
 from dipdup.exceptions import ConfigurationError
 from dipdup.exceptions import FrameworkException
 from dipdup.exceptions import HasuraError
@@ -28,8 +32,6 @@ from dipdup.http import HTTPGateway
 from dipdup.models import Schema
 from dipdup.utils import iter_files
 from dipdup.utils import pascal_to_snake
-from dipdup.utils.database import get_connection
-from dipdup.utils.database import iter_models
 
 vulnerable_versions = {
     # NOTE: See https://github.com/hasura/graphql-engine/security/advisories/GHSA-c9rw-rw2f-mj4x
@@ -59,10 +61,8 @@ vulnerable_versions = {
     'v1.3.0': 'v1.3.4',
 }
 
-RelationalFieldT = Union[
-    fields.relational.ForeignKeyFieldInstance,
-    fields.relational.ManyToManyFieldInstance,
-]
+RelationalFieldT = fields.relational.ForeignKeyFieldInstance[Any] | fields.relational.ManyToManyFieldInstance[Any]
+
 _get_fields_query = """
 query introspectionQuery($name: String!) {
   __type(name: $name) {
@@ -82,11 +82,7 @@ query introspectionQuery($name: String!) {
     }
   }
 }
-""".replace(
-    '\n', ' '
-).replace(
-    '  ', ''
-)
+""".replace('\n', ' ').replace('  ', '')
 
 
 @dataclass
@@ -106,12 +102,14 @@ class Field:
 
 
 class HasuraGateway(HTTPGateway):
-    _default_http_config = HTTPConfig(
+    _default_http_config = HttpConfig(
         # NOTE: Fail fast; most Hasura errors are 500's that won't fix by themselves.
         # NOTE: Does not apply to initial healthcheck
         retry_sleep=1,
         retry_multiplier=1.1,
         retry_count=3,
+        alias='hasura',
+        replay=False,
     )
 
     def __init__(
@@ -119,13 +117,13 @@ class HasuraGateway(HTTPGateway):
         package: str,
         hasura_config: HasuraConfig,
         database_config: PostgresDatabaseConfig,
-        http_config: HTTPConfig | None = None,
+        http_config: HttpConfig | None = None,
     ) -> None:
         super().__init__(
             hasura_config.url,
-            ResolvedHTTPConfig.create(self._default_http_config, http_config),
+            ResolvedHttpConfig.create(self._default_http_config, http_config),
         )
-        self._logger = logging.getLogger('dipdup.hasura')
+        self._logger = logging.getLogger(__name__)
         self._package = package
         self._hasura_config = hasura_config
         self._database_config = database_config
@@ -196,6 +194,22 @@ class HasuraGateway(HTTPGateway):
 
         await self._apply_custom_metadata()
 
+        # Apply model docstrings as database comments
+        await self._apply_model_comments()
+
+        # Force metadata refresh to pick up new database comments
+        self._logger.info('Refreshing Hasura metadata to pick up new database comments')
+        await self._hasura_request(
+            endpoint='metadata',
+            json={
+                'type': 'reload_metadata',
+                'args': {
+                    'reload_remote_schemas': True,
+                    'recreate_event_triggers': True,
+                },
+            },
+        )
+
         # TODO: Find out why it is necessary
         # NOTE: Fetch metadata once again and save its hash for future comparisons
         metadata = await self._fetch_metadata()
@@ -208,13 +222,20 @@ class HasuraGateway(HTTPGateway):
     def _get_source(self, metadata: dict[str, Any], name: str) -> dict[str, Any] | None:
         for source in metadata['sources']:
             if source['name'] == name:
-                return cast(dict[str, Any], source)
+                return cast('dict[str, Any]', source)
         else:
             return None
 
-    async def _hasura_request(self, endpoint: str, json: dict[str, Any] | None = None) -> dict[str, Any]:
+    async def _hasura_request(
+        self,
+        endpoint: str,
+        json: dict[str, Any] | None = None,
+        retry_count: int | None = None,
+    ) -> dict[str, Any]:
         self._logger.debug('Sending `%s` request: %s', endpoint, orjson.dumps(json))
         try:
+            if retry_count is not None:
+                self._http_config.retry_count, retry_count = retry_count, self._http_config.retry_count
             result = await self.request(
                 method='get' if json is None else 'post',
                 url=f'v1/{endpoint}',
@@ -223,16 +244,22 @@ class HasuraGateway(HTTPGateway):
             )
         except ClientResponseError as e:
             raise HasuraError(f'{e.status} {e.message}') from e
+        finally:
+            if retry_count is not None:
+                self._http_config.retry_count, retry_count = retry_count, self._http_config.retry_count
 
         self._logger.debug('Response: %s', result)
+        if isinstance(result, list):
+            # for bulk requests with response code 200 it's always list of {"message":"success"}
+            result = result[0]
         if errors := result.get('error') or result.get('errors'):
             raise HasuraError(errors)
 
-        return cast(dict[str, Any], result)
+        return cast('dict[str, Any]', result)
 
     async def _healthcheck(self) -> None:
         self._logger.info('Connecting to Hasura instance')
-        version_json = await self._hasura_request('version')
+        version_json = await self._hasura_request('version', retry_count=20)
         version = version_json['version']
         if version.startswith('v1'):
             raise UnsupportedAPIError(
@@ -244,13 +271,16 @@ class HasuraGateway(HTTPGateway):
             raise UnsupportedAPIError(
                 self.url,
                 'hasura',
-                f'A critical vulnerability has been discovered in {version}!\n    Update to {vulnerable_versions[version]} or the latest stable version immediately.',
+                (
+                    f'A critical vulnerability has been discovered in {version}!\n    Update to'
+                    f' {vulnerable_versions[version]} or the latest stable version immediately.'
+                ),
             )
 
         self._logger.info('Connected to Hasura %s', version)
 
     async def _create_source(self) -> dict[str, Any]:
-        self._logger.info(f'Adding source `{self._hasura_config.source}`')
+        self._logger.info('Adding source `%s`', self._hasura_config.source)
         return await self._hasura_request(
             endpoint='metadata',
             json={
@@ -272,12 +302,14 @@ class HasuraGateway(HTTPGateway):
 
     async def _fetch_metadata(self) -> dict[str, Any]:
         self._logger.info('Fetching existing metadata')
-        return await self._hasura_request(
-            endpoint='metadata',
-            json={
-                'type': 'export_metadata',
-                'args': {},
-            },
+        return dict(
+            await self._hasura_request(
+                endpoint='metadata',
+                json={
+                    'type': 'export_metadata',
+                    'args': {},
+                },
+            )
         )
 
     def _hash_metadata(self, metadata: dict[str, Any]) -> str:
@@ -315,23 +347,22 @@ class HasuraGateway(HTTPGateway):
 
     async def _get_views(self) -> list[str]:
         conn = get_connection()
-        views = [
-            row[0]
-            for row in (
-                await conn.execute_query(
-                    f"SELECT table_name FROM information_schema.views WHERE table_schema = '{self._database_config.schema_name}' UNION "
-                    f"SELECT matviewname as table_name FROM pg_matviews WHERE schemaname = '{self._database_config.schema_name}'"
-                )
-            )[1]
-        ]
-        self._logger.info('Found %s regular and materialized views', len(views))
-        return views
+        if not isinstance(conn, AsyncpgClient):
+            raise HasuraError('Hasura integration requires `postgres` database client')
+        return await pg_get_views(conn, self._database_config.schema_name)
 
     def _iterate_graphql_queries(self) -> Iterator[tuple[str, str]]:
         graphql_path = env.get_package_path(self._package) / 'graphql'
         for file in iter_files(graphql_path, '.graphql'):
             yield file.name.split('/')[-1][:-8], file.read()
 
+    def _is_hidden(self, name: str) -> bool:
+        if self._hasura_config.hide_internal and name.startswith('dipdup_'):
+            return True
+        if name in self._hasura_config.hide:
+            return True
+        return False
+
     async def _generate_source_tables_metadata(self) -> list[dict[str, Any]]:
         """Generate source tables metadata based on project models and views.
 
@@ -353,11 +384,16 @@ class HasuraGateway(HTTPGateway):
             metadata_tables[view] = self._format_table(view)
 
         for app, model in iter_models(self._package):
-            table_name = model_tables[f'{app}.{model.__name__}']
+            table_name = model_tables.get(f'{app}.{model.__name__}')  # type: ignore[assignment]
+            if not table_name:
+                continue
 
             for field in model._meta.fields_map.values():
                 if isinstance(field, fields.relational.ForeignKeyFieldInstance):
-                    related_table_name = model_tables[field.model_name]
+                    related_table_name = model_tables.get(field.model_name)
+                    if not related_table_name:
+                        continue
+
                     field_name = field.model_field_name
                     metadata_tables[table_name]['object_relationships'].append(
                         self._format_object_relationship(
@@ -375,7 +411,10 @@ class HasuraGateway(HTTPGateway):
                         )
 
                 elif isinstance(field, fields.relational.ManyToManyFieldInstance):
-                    related_table_name = model_tables[field.model_name]
+                    related_table_name = model_tables.get(field.model_name)
+                    if not related_table_name:
+                        continue
+
                     junction_table_name = field.through
 
                     metadata_tables[junction_table_name] = self._format_table(junction_table_name)
@@ -431,7 +470,7 @@ class HasuraGateway(HTTPGateway):
             queries.append({'name': query_name, 'query': query})
 
         # NOTE: This is the only view we add by ourselves and thus know all params. Won't work for any view.
-        queries.append(self._format_rest_head_status_query())
+        queries.append(self._format_rest_status_query())
 
         return queries
 
@@ -450,7 +489,7 @@ class HasuraGateway(HTTPGateway):
             },
         )
         try:
-            return cast(list[dict[str, Any]], result['data']['__type']['fields'])
+            return cast('list[dict[str, Any]]', result['data']['__type']['fields'])
         except TypeError as e:
             raise HasuraError(f'Unknown table `{name}`') from e
 
@@ -501,7 +540,7 @@ class HasuraGateway(HTTPGateway):
         table_names = {t['table']['name'] for t in source['tables']}
         tables = await self._get_fields()
 
-        # TODO: Use new Hasura API
+        bulk_request_args: list[dict[str, Any]] = []
         self._logger.info('Applying table customizations (could take some time)')
         for table in tables:
             if table.root not in table_names:
@@ -509,24 +548,27 @@ class HasuraGateway(HTTPGateway):
 
             custom_root_fields = self._format_custom_root_fields(table.root)
             columns = await self._get_fields(table.root)
-            custom_column_names = self._format_custom_column_names(columns)
+            column_config = self._format_column_config(columns)
             args: dict[str, Any] = {
                 'table': self._format_table_table(table.root),
                 'source': self._hasura_config.source,
                 'configuration': {
                     'identifier': custom_root_fields['select_by_pk'],
                     'custom_root_fields': custom_root_fields,
-                    'custom_column_names': custom_column_names,
+                    'column_config': column_config,
                 },
             }
 
-            await self._hasura_request(
-                endpoint='metadata',
-                json={
-                    'type': 'pg_set_table_customization',
-                    'args': args,
-                },
-            )
+            bulk_request_args.append({'type': 'pg_set_table_customization', 'args': args})
+
+        await self._hasura_request(
+            endpoint='metadata',
+            json={
+                'type': 'bulk',
+                'source': self._hasura_config.source,
+                'args': bulk_request_args,
+            },
+        )
 
     def _format_rest_query(self, name: str, table: str, filter: str, fields: Iterable[Field]) -> dict[str, Any]:
         if not table.endswith('_by_pk'):
@@ -548,27 +590,21 @@ class HasuraGateway(HTTPGateway):
         query_fields = ' '.join(f.name for f in fields)
         return {
             'name': name,
-            'query': 'query '
-            + name
-            + ' ('
-            + query_arg
-            + ') {'
-            + table
-            + '('
-            + query_filter
-            + ') {'
-            + query_fields
-            + '}}',
+            'query': (
+                'query ' + name + ' (' + query_arg + ') {' + table + '(' + query_filter + ') {' + query_fields + '}}'
+            ),
         }
 
-    def _format_rest_head_status_query(self) -> dict[str, Any]:
-        name = 'dipdup_head_status'
+    def _format_rest_status_query(self) -> dict[str, Any]:
+        name = 'dipdup_status'
+        fields = '{type name level size updated_at}'
         if self._hasura_config.camel_case:
             name = humps.camelize(name)
+            fields = fields.replace('updated_at', 'updatedAt')
 
         return {
             'name': name,
-            'query': 'query ' + name + ' ($name: String!) {' + name + '(where: {name: {_eq: $name}}) {status}}',
+            'query': 'query ' + name + ' ($name: String!) {' + name + '(where: {name: {_eq: $name}}) ' + fields + '}',
         }
 
     def _format_rest_endpoint(self, query_name: str) -> dict[str, Any]:
@@ -605,19 +641,26 @@ class HasuraGateway(HTTPGateway):
         }
 
     def _format_custom_column_names(self, fields: list[Field]) -> dict[str, Any]:
+        """
+        Deprecated
+        See: https://hasura.io/docs/2.0/api-reference/syntax-defs/#customcolumnnames
+        """
         if self._hasura_config.camel_case:
             return {humps.decamelize(f.name): humps.camelize(f.name) for f in fields}
-        else:
-            return {humps.decamelize(f.name): humps.decamelize(f.name) for f in fields}
+        return {humps.decamelize(f.name): humps.decamelize(f.name) for f in fields}
+
+    def _format_column_config(self, fields: list[Field]) -> dict[str, Any]:
+        if self._hasura_config.camel_case:
+            return {humps.decamelize(f.name): {'custom_name': humps.camelize(f.name)} for f in fields}
+        return {humps.decamelize(f.name): {'custom_name': humps.decamelize(f.name)} for f in fields}
 
     def _format_table(self, name: str) -> dict[str, Any]:
+        permissions = [] if self._is_hidden(name) else [self._format_select_permissions()]
         return {
             'table': self._format_table_table(name),
             'object_relationships': [],
             'array_relationships': [],
-            'select_permissions': [
-                self._format_select_permissions(),
-            ],
+            'select_permissions': permissions,
         }
 
     def _format_table_table(self, name: str) -> dict[str, Any]:
@@ -671,5 +714,129 @@ class HasuraGateway(HTTPGateway):
 
     def _iterate_metadata_requests(self) -> Iterator[TextIO]:
         metadata_path = env.get_package_path(self._package) / 'hasura'
-        for file in iter_files(metadata_path, '.json'):
-            yield file
+        yield from iter_files(metadata_path, '.json')
+
+    async def _apply_model_comments(self) -> None:
+        """Extract docstrings from Tortoise models and apply them as database comments.
+
+        This method scans all models in the project and extracts their docstrings,
+        then applies them as SQL comments to the corresponding database tables and columns.
+        These comments will be visible in Hasura's GraphQL playground.
+        """
+        self._logger.info('Extracting model docstrings and applying database comments')
+
+        # Get database connection
+        conn = get_connection()
+        if not isinstance(conn, AsyncpgClient):
+            self._logger.warning('Model comments only supported for PostgreSQL databases')
+            return
+
+        # Debug: Log the package being processed
+        self._logger.debug('Processing package: %s', self._package)
+
+        # Iterate through all models
+        model_count = 0
+        for app_name, model_class in iter_models(self._package):
+            model_count += 1
+            self._logger.debug('Processing model %s: %s from %s', model_count, model_class.__name__, app_name)
+
+            if not hasattr(model_class, '_meta') or not hasattr(model_class._meta, 'db_table'):
+                self._logger.warning('Model %s missing _meta or db_table', model_class.__name__)
+                continue
+
+            table_name = model_class._meta.db_table
+            self._logger.debug('Table name: %s', table_name)
+
+            # Extract table-level docstring
+            table_doc = model_class.__doc__ or ''
+            if table_doc:
+                # Clean up the docstring
+                table_doc = table_doc.strip().replace("'", "''")  # Escape single quotes
+                if len(table_doc) > 1000:  # Limit comment length
+                    table_doc = table_doc[:997] + '...'
+
+                # Apply table comment
+                try:
+                    sql = f"COMMENT ON TABLE {self._database_config.schema_name}.{table_name} IS '{table_doc}';"
+                    self._logger.debug('Executing table comment SQL: %s', sql)
+                    await conn.execute_script(sql)
+                    self._logger.debug('Successfully applied table comment to %s', table_name)
+                except Exception as e:
+                    self._logger.warning('Failed to apply table comment to %s: %s', table_name, e)
+            else:
+                self._logger.debug('No table docstring found for %s', table_name)
+
+            # Extract field-level comments from model source code
+            # Since Tortoise doesn't natively support field descriptions, we'll look for
+            # comments in the model definition using inspect
+            try:
+                source_lines = inspect.getsource(model_class).split('\n')
+                field_comments = {}
+
+                self._logger.debug('Extracting field comments from %s, %s lines', table_name, len(source_lines))
+
+                for i, line in enumerate(source_lines):
+                    line = line.strip()
+                    # Look for field definitions - check for both quoted and unquoted field references
+                    if (
+                        '=' in line
+                        and ('fields.' in line or "'fields." in line or '"fields.' in line)
+                        and not line.startswith('#')
+                    ):
+                        # Extract field name
+                        field_name = line.split('=')[0].strip()
+                        if field_name and not field_name.startswith('_'):
+                            # Look for comments above this line
+                            comment_lines: list[str] = []
+                            j = i - 1
+                            while j >= 0:
+                                prev_line = source_lines[j].strip()
+                                if prev_line.startswith('#'):
+                                    comment_lines.insert(0, prev_line[1:].strip())
+                                    j -= 1
+                                elif prev_line == '':
+                                    j -= 1
+                                else:
+                                    break
+
+                            if comment_lines:
+                                field_comments[field_name] = ' '.join(comment_lines)
+                                self._logger.debug(
+                                    'Found comment for field %s: %s', field_name, field_comments[field_name]
+                                )
+
+                self._logger.debug('Found %s field comments for %s', len(field_comments), table_name)
+
+                # Apply field comments
+                if hasattr(model_class._meta, 'fields_map'):
+                    self._logger.debug('Processing %s fields for %s', len(model_class._meta.fields_map), table_name)
+                    for field_name, _field in model_class._meta.fields_map.items():
+                        self._logger.debug('Checking field: %s', field_name)
+                        if field_name in field_comments:
+                            field_doc = field_comments[field_name].strip().replace("'", "''")
+                            if len(field_doc) > 1000:
+                                field_doc = field_doc[:997] + '...'
+
+                            # Get the actual database column name
+                            db_column = model_class._meta.fields_db_projection.get(field_name, field_name)
+
+                            try:
+                                sql = f"COMMENT ON COLUMN {self._database_config.schema_name}.{table_name}.{db_column} IS '{field_doc}';"
+                                self._logger.debug('Executing column comment SQL: %s', sql)
+                                await conn.execute_script(sql)
+                                self._logger.debug(
+                                    'Successfully applied column comment to %s.%s', table_name, db_column
+                                )
+                            except Exception as e:
+                                self._logger.warning(
+                                    'Failed to apply column comment to %s.%s: %s', table_name, db_column, e
+                                )
+                        else:
+                            self._logger.debug('No comment found for field %s in %s', field_name, table_name)
+                else:
+                    self._logger.warning('No fields_map found for %s', table_name)
+
+            except Exception as e:
+                self._logger.warning('Failed to extract field comments from %s: %s', table_name, e)
+
+        self._logger.debug('Finished applying model comments to database. Processed %s models.', model_count)
