@@ -2,32 +2,33 @@ import asyncio
 import hashlib
 import logging
 import platform
-import sys
+import time
 from contextlib import AbstractAsyncContextManager
 from contextlib import suppress
 from http import HTTPStatus
 from json import JSONDecodeError
 from pathlib import Path
+from random import random
 from typing import Any
-from typing import Mapping
-from typing import Optional
-from typing import cast
+from typing import Literal
+from typing import overload
+from urllib.parse import parse_qsl
 from urllib.parse import urlsplit
 from urllib.parse import urlunsplit
 
 import aiohttp
-import aiohttp.test_utils
 import orjson
 from aiolimiter import AsyncLimiter
 
 from dipdup import __version__
-from dipdup.config import HTTPConfig
-from dipdup.config import ResolvedHTTPConfig
+from dipdup import env
+from dipdup.config import ResolvedHttpConfig
 from dipdup.exceptions import FrameworkException
 from dipdup.exceptions import InvalidRequestError
-from dipdup.prometheus import Metrics
+from dipdup.performance import metrics
 
 safe_exceptions = (
+    asyncio.TimeoutError,
     aiohttp.ClientConnectionError,
     aiohttp.ClientConnectorError,
     aiohttp.ClientResponseError,
@@ -38,9 +39,7 @@ safe_exceptions = (
 class HTTPGateway(AbstractAsyncContextManager[None]):
     """Base class for datasources which connect to remote HTTP endpoints"""
 
-    _default_http_config: HTTPConfig
-
-    def __init__(self, url: str, http_config: ResolvedHTTPConfig) -> None:
+    def __init__(self, url: str, http_config: ResolvedHttpConfig) -> None:
         self._http_config = http_config
         self._http = _HTTPGateway(url, self._http_config)
 
@@ -55,7 +54,7 @@ class HTTPGateway(AbstractAsyncContextManager[None]):
     @property
     def url(self) -> str:
         """HTTP endpoint URL"""
-        return self._http._url
+        return self._http._url + self._http._path
 
     async def request(
         self,
@@ -77,28 +76,33 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
 
     Covers caching, retrying failed requests and ratelimiting"""
 
-    def __init__(self, url: str, config: ResolvedHTTPConfig) -> None:
-        self._logger = logging.getLogger('dipdup.http')
+    def __init__(self, url: str, config: ResolvedHttpConfig) -> None:
+        self._logger = logging.getLogger(__name__)
         parsed_url = urlsplit(url)
         self._url = urlunsplit((parsed_url.scheme, parsed_url.netloc, '', '', ''))
+        self._alias = config.alias or parsed_url.netloc
         self._path = parsed_url.path
+        self._query = parsed_url.query
         self._config = config
         self._user_agent_args: tuple[str, ...] = ()
-        self._user_agent: Optional[str] = None
+        self._user_agent: str | None = None
         self._ratelimiter = (
             AsyncLimiter(max_rate=config.ratelimit_rate, time_period=config.ratelimit_period)
             if config.ratelimit_rate and config.ratelimit_period
             else None
         )
-        self.__session: Optional[aiohttp.ClientSession] = None
+        self.__session: aiohttp.ClientSession | None = None
 
     async def __aenter__(self) -> None:
         """Create underlying aiohttp session"""
         self.__session = aiohttp.ClientSession(
             base_url=self._url,
-            json_serialize=lambda *a, **kw: orjson.dumps(*a, **kw).decode(),
+            json_serialize=lambda v: orjson.dumps(v).decode(),
             connector=aiohttp.TCPConnector(limit=self._config.connection_limit),
-            timeout=aiohttp.ClientTimeout(connect=self._config.connection_timeout),
+            timeout=aiohttp.ClientTimeout(
+                total=self._config.request_timeout,
+                connect=self._config.connection_timeout,
+            ),
         )
 
     async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None:
@@ -121,6 +125,8 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
     @property
     def _session(self) -> aiohttp.ClientSession:
         """Get an aiohttp session from inside of it's context manager"""
+        import aiohttp.test_utils
+
         if isinstance(self.__session, aiohttp.test_utils.TestClient):
             return self.__session
         if self.__session is None:
@@ -141,10 +147,11 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
         attempt = 1
         retry_sleep = self._config.retry_sleep
         retry_count = self._config.retry_count
-        retry_count_str = 'inf' if retry_count is sys.maxsize else str(retry_count)
+        last_attempt = retry_count + 1
+
+        metrics.set_http_errors_in_row(self._url, 0)
 
         while True:
-            self._logger.debug('HTTP request attempt %s/%s', attempt, retry_count_str)
             try:
                 return await self._request(
                     method=method,
@@ -153,44 +160,78 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
                     **kwargs,
                 )
             except safe_exceptions as e:
-                if self._config.retry_count and attempt - 1 == self._config.retry_count:
-                    raise e
-
                 ratelimit_sleep: float | None = None
                 if isinstance(e, aiohttp.ClientResponseError):
-                    if Metrics.enabled:
-                        Metrics.set_http_error(self._url, e.status)
+                    metrics.set_http_error(self._url, e.status)
 
                     if e.status == HTTPStatus.TOO_MANY_REQUESTS:
                         ratelimit_sleep = self._config.ratelimit_sleep
                         # TODO: Parse Retry-After in UTC date format
                         with suppress(KeyError, ValueError):
-                            e.headers = cast(Mapping[str, Any], e.headers)
-                            ratelimit_sleep = max(ratelimit_sleep, int(e.headers['Retry-After']))
+                            ratelimit_sleep = max(ratelimit_sleep, int(e.headers['Retry-After']))  # type: ignore[index]
+                        # randomize to avoid thundering herd
+                        ratelimit_sleep *= 1 + (random() - 0.5) / 5
                 else:
-                    if Metrics.enabled:
-                        Metrics.set_http_error(self._url, 0)
+                    metrics.set_http_error(self._url, 0)
+
+                self._logger.warning('HTTP request attempt %s/%s failed: %s', attempt, last_attempt, e)
+                metrics.set_http_errors_in_row(self._url, attempt)
+                if attempt == last_attempt:
+                    raise e
+
+                to_sleep = float(f'{ratelimit_sleep or retry_sleep:.1f}')
+                self._logger.info('Waiting %s seconds before retry', to_sleep)
+                await asyncio.sleep(to_sleep)
 
-                self._logger.warning('HTTP request attempt %s/%s failed: %s', attempt, retry_count_str, e)
-                self._logger.info('Waiting %s seconds before retry', ratelimit_sleep or retry_sleep)
-                await asyncio.sleep(ratelimit_sleep or retry_sleep)
                 attempt += 1
                 if not ratelimit_sleep:
                     retry_sleep *= self._config.retry_multiplier
 
+    # FIXME: Temporary overload for Subsquid; move to public method
+    @overload
+    async def _request(
+        self,
+        method: str,
+        url: str,
+        weight: int,
+        raw: Literal[True],
+        **kwargs: Any,
+    ) -> aiohttp.ClientResponse: ...
+
+    @overload
+    async def _request(
+        self,
+        method: str,
+        url: str,
+        weight: int,
+        raw: Literal[False],
+        **kwargs: Any,
+    ) -> Any: ...
+
     async def _request(
         self,
         method: str,
         url: str,
         weight: int = 1,
+        raw: bool = False,
         **kwargs: Any,
     ) -> Any:
         """Wrapped aiohttp call with preconfigured headers and ratelimiting"""
-        url = f"{self._path.rstrip('/')}/{url}"
+        metrics.requests_total[self._alias] += 1
+        url = url or '/'
+
+        if url.startswith('http'):
+            raise InvalidRequestError(msg='URL should not start with http(s)://', url=url)
+
+        url = f'{self._path.strip("/")}/{url.strip("/")}'.rstrip('/')
+
         headers = kwargs.pop('headers', {})
         headers['User-Agent'] = self.user_agent
 
-        params = kwargs.get('params', {})
+        params = kwargs.pop('params', {})
+        for item in parse_qsl(self._query):
+            params[item[0]] = item[1]
+
         params_string = '&'.join(f'{k}={v}' for k, v in params.items())
         request_string = f'{self._url}{url}?{params_string}'.rstrip('?')
         self._logger.debug('Calling `%s`', request_string)
@@ -198,18 +239,29 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
         if self._ratelimiter:
             await self._ratelimiter.acquire(weight)
 
+        started_at = time.time()
+
         async with self._session.request(
             method=method,
             url=url,
             headers=headers,
-            raise_for_status=True,
+            raise_for_status=not raw,
+            params=params,
             **kwargs,
         ) as response:
+            await response.read()
+            metrics.time_in_requests[self._alias] += time.time() - started_at
+            if raw:
+                return response
+
+            # NOTE: Use raw=True if fail on 204 is not a desired behavior
             if response.status == HTTPStatus.NO_CONTENT:
                 raise InvalidRequestError('204 No Content', request_string)
-            with suppress(JSONDecodeError, aiohttp.ContentTypeError):
-                return await response.json()
-            return await response.read()
+            if not response._body:
+                raise InvalidRequestError('Empty response', request_string)
+            with suppress(JSONDecodeError):
+                return orjson.loads(response._body)
+            return response._body
 
     async def _replay_request(
         self,
@@ -218,16 +270,20 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
         weight: int = 1,
         **kwargs: Any,
     ) -> Any:
-        if not self._config.replay_path:
-            raise FrameworkException('Replay path is not set')
-
-        replay_path = Path(self._config.replay_path).expanduser()
+        replay_path = self._config.replay_path or env.REPLAY_PATH or Path('~/.cache/dipdup/replays')
+        if isinstance(replay_path, str):
+            replay_path = Path(replay_path)
         replay_path.mkdir(parents=True, exist_ok=True)
 
-        request_hash = hashlib.sha256(
-            f'{self._url} {method} {self._path}/{url} {kwargs}'.encode(),
-        ).hexdigest()
-        replay_path = Path(self._config.replay_path).joinpath(request_hash).expanduser()
+        hashable_kwargs = {**kwargs}
+        # NOTE: JSONRPC fix
+        if (json := hashable_kwargs.get('json')) and 'jsonrpc' in json:
+            json['id'] = ''
+
+        # FIXME: Fails for
+        request_repr = f'{self._url} {method} {self._path}/{url} {hashable_kwargs}'
+        request_hash = hashlib.sha256(request_repr.encode()).hexdigest()
+        replay_path = replay_path.joinpath(request_hash).expanduser()
 
         if replay_path.exists():
             if not replay_path.stat().st_size:
@@ -235,7 +291,12 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
 
             content = replay_path.read_bytes()
             with suppress(JSONDecodeError):
-                return orjson.loads(content)
+                response = orjson.loads(content)
+                # NOTE: JSONRPC fix
+                if isinstance(response, dict) and 'jsonrpc' in response:
+                    response['id'] = kwargs['json']['id']
+                return response
+
             return content
 
         response = await self._retry_request(method, url, weight, **kwargs)
@@ -255,10 +316,9 @@ class _HTTPGateway(AbstractAsyncContextManager[None]):
         **kwargs: Any,
     ) -> Any:
         """Performs an HTTP request."""
-        if self._config.replay_path:
+        if self._config.replay_path or self._config.replay is True or (self._config.replay is None and env.REPLAY_PATH):
             return await self._replay_request(method, url, weight, **kwargs)
-        else:
-            return await self._retry_request(method, url, weight, **kwargs)
+        return await self._retry_request(method, url, weight, **kwargs)
 
     def set_user_agent(self, *args: str) -> None:
         """Add list of arguments to User-Agent header"""
