@@ -1,20 +1,16 @@
 import asyncio
-import json
 import logging
 from contextlib import suppress
 from functools import partial
 from typing import Any
-from typing import Dict
-from typing import Optional
-from typing import Set
 
-from apscheduler.events import EVENT_JOB_ERROR  # type: ignore[import]
+from apscheduler.events import EVENT_JOB_ERROR  # type: ignore[import-untyped]
 from apscheduler.events import EVENT_JOB_EXECUTED
 from apscheduler.events import JobEvent
-from apscheduler.job import Job  # type: ignore[import]
-from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore[import]
-from apscheduler.triggers.cron import CronTrigger  # type: ignore[import]
-from apscheduler.triggers.interval import IntervalTrigger  # type: ignore[import]
+from apscheduler.job import Job  # type: ignore[import-untyped]
+from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore[import-untyped]
+from apscheduler.triggers.cron import CronTrigger  # type: ignore[import-untyped]
+from apscheduler.triggers.interval import IntervalTrigger  # type: ignore[import-untyped]
 
 from dipdup.config import JobConfig
 from dipdup.context import DipDupContext
@@ -22,6 +18,7 @@ from dipdup.context import HookContext
 from dipdup.exceptions import ConfigurationError
 from dipdup.exceptions import FrameworkException
 from dipdup.utils import FormattedLogger
+from dipdup.utils import json_dumps
 
 DEFAULT_CONFIG = {
     'apscheduler.jobstores.default.class': 'apscheduler.jobstores.memory:MemoryJobStore',
@@ -30,9 +27,9 @@ DEFAULT_CONFIG = {
 }
 
 
-def _verify_config(config: Dict[str, Any]) -> None:
+def _verify_config(config: dict[str, Any]) -> None:
     """Ensure that dict is a valid `apscheduler` config"""
-    json_config = json.dumps(config)
+    json_config = json_dumps(config).decode()
     if 'apscheduler.executors.pool' in json_config:
         raise ConfigurationError(
             '`apscheduler.executors.pool` is not supported. If needed, create a pool inside a regular hook.'
@@ -45,24 +42,34 @@ def _verify_config(config: Dict[str, Any]) -> None:
 
 
 class SchedulerManager:
-    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
+    def __init__(
+        self,
+        jobs: dict[str, JobConfig],
+        config: dict[str, Any] | None = None,
+    ) -> None:
         if config:
             _verify_config(config)
-        self._logger = logging.getLogger('dipdup.jobs')
+        self._logger = logging.getLogger(__name__)
+        self._jobs = jobs
         self._scheduler = AsyncIOScheduler(config or DEFAULT_CONFIG)
         self._scheduler.add_listener(self._on_error, EVENT_JOB_ERROR)
         self._scheduler.add_listener(self._on_executed, EVENT_JOB_EXECUTED)
-        self._exception: Optional[Exception] = None
+        self._exception: Exception | None = None
         self._exception_event: asyncio.Event = asyncio.Event()
-        self._daemons: Set[str] = set()
+        self._daemons: set[str] = set()
 
-    async def run(self, event: asyncio.Event) -> None:
-        self._logger.info('Waiting for an event to start scheduler')
+    async def run(self, ctx: 'DipDupContext', event: asyncio.Event) -> None:
+        if not event.is_set():
+            self._logger.info('Job scheduler is waiting for an event')
         await event.wait()
 
-        self._logger.info('Starting scheduler')
         try:
+            self._logger.info('Starting job scheduler')
             self._scheduler.start()
+
+            for job_config in self._jobs.values():
+                self.add_job(ctx, job_config)
+
             await self._exception_event.wait()
             if self._exception is None:
                 raise FrameworkException('Job has failed but exception is not set')
@@ -72,7 +79,7 @@ class SchedulerManager:
         finally:
             self._scheduler.shutdown()
 
-    def add_job(self, ctx: DipDupContext, job_config: JobConfig) -> Job:
+    def add_job(self, ctx: 'DipDupContext', job_config: JobConfig) -> Job:
         if job_config.daemon:
             self._daemons.add(job_config.name)
 
@@ -84,23 +91,19 @@ class SchedulerManager:
         )
 
         async def _job_wrapper(
-            ctx: DipDupContext,
-            *args: Any,
+            ctx: 'DipDupContext',
             **kwargs: Any,
         ) -> None:
             nonlocal job_config, hook_config
-            job_ctx = HookContext(
-                config=ctx.config,
-                datasources=ctx.datasources,
-                callbacks=ctx._callbacks,
-                transactions=ctx._transactions,
+            job_ctx = HookContext._wrap(
+                ctx=ctx,
                 logger=logger,
                 hook_config=hook_config,
             )
             with suppress(asyncio.CancelledError):
                 await job_ctx.fire_hook(
-                    hook_config.callback,
-                    *args,
+                    name=hook_config.callback,
+                    wait=True,
                     **kwargs,
                 )
 
