@@ -1,14 +1,14 @@
 from collections import deque
+from collections.abc import AsyncIterator
 from contextlib import asynccontextmanager
-from typing import AsyncIterator
-from typing import Optional
-from typing import Set
 
 from tortoise.transactions import in_transaction
 
 import dipdup.models
-from dipdup.utils.database import get_connection
-from dipdup.utils.database import set_connection
+from dipdup.config import WatchdogTrigger
+from dipdup.database import get_connection
+from dipdup.database import set_connection
+from dipdup.watchdog import watchdog
 
 
 class TransactionManager:
@@ -16,12 +16,12 @@ class TransactionManager:
 
     def __init__(
         self,
-        depth: int = 2,
-        immune_tables: Optional[Set[str]] = None,
+        depth: int | None = None,
+        immune_tables: set[str] | None = None,
     ) -> None:
         self._depth = depth
         self._immune_tables = immune_tables or set()
-        self._transaction: Optional[dipdup.models.VersionedTransaction] = None
+        self._transaction: dipdup.models.VersionedTransaction | None = None
         self._pending_updates: deque[dipdup.models.ModelUpdate] = deque()
 
     @asynccontextmanager
@@ -39,12 +39,13 @@ class TransactionManager:
     @asynccontextmanager
     async def in_transaction(
         self,
-        level: Optional[int] = None,
-        sync_level: Optional[int] = None,
-        index: Optional[str] = None,
+        level: int | None = None,
+        sync_level: int | None = None,
+        index: str | None = None,
     ) -> AsyncIterator[None]:
         """Enforce using transaction for all queries inside wrapped block. Works for a single DB only."""
         try:
+            watchdog.heartbeat(WatchdogTrigger.transaction)
             original_conn = get_connection()
             async with in_transaction() as conn:
                 set_connection(conn)
@@ -67,6 +68,7 @@ class TransactionManager:
         finally:
             self._transaction = None
             set_connection(original_conn)
+            watchdog.reset(WatchdogTrigger.transaction)
 
     async def _commit(self) -> None:
         """Save pending updates to DB in the same order as they were added"""
@@ -75,6 +77,8 @@ class TransactionManager:
 
     async def cleanup(self) -> None:
         """Cleanup outdated model updates"""
+        if not self._depth:
+            return
         most_recent_index = await dipdup.models.Index.filter().order_by('-level').first()
         if not most_recent_index:
             return
